# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rucio.rse.protocols.protocol import RSEDeterministicTranslation


def lfn2pfn_module_algorithm(scope, name, rse, rse_attributes, protocol_attributes):
    """Dummy LFN2PFN algorithm for unit tests."""
    del scope
    del name
    del rse
    del rse_attributes
    del protocol_attributes
    return "lfn2pfn_module_algorithm_value"


RSEDeterministicTranslation.register(lfn2pfn_module_algorithm)
