
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.29.9',
    'branch_nick': 'release-1.29-LTS',
    'revision_id': 'e63c74ba171b04b0e838475d6ade4169af7538af',
    'revno': 12094
}
